#ifndef POPTALENV_H
#define POPTALENV_H

#include "core.h"
#include "xarray.h"
#include "dict.h"
#include "id.h"
#include "set.h"
#include "bitvec.h"

#include "poperr.h"
#include "popil.h"
#include "tal.h"

prefix Poptalenv {
open   Poptalenv {

open Core {
open Popil {
open Tal {

#define list   List::list
#define id     Id::id
#define dict   Dict::dict
#define xarray Xarray::xarray

extern struct reg_assignment {
  // what the register allocator passes to the code generator
  tal_place     assignment[]; // the actual assignment to regs and slots
  <int>Set::set all_live[][]; // needed for TAL label types
  int           num_slots;    // number of stack slots used for spills
  int           names[];      // names[i] is TAL name to use for variable i
                              // (aliases must use same name)
}

// This environment is for the term translation
extern struct cg_env {
  // later put stuff in here for type-sharing (or for prototype do as post pass)
  // probably need two-way map between cons and abbrevs
  
  // this stuff only changes once per function 
    // set by caller of trans_function
  cf_function    fn; 
  reg_assignment assignment_info;
    // set by trans_function
  con            stack_bottom; // call stack_on_entry to set this,
                               // does NOT include handler and below
  con            args; // like stack_bottom, but more convenient for new abbrevs
  <int,con>dict  handler_types; // block num to label type (handlers only)
  bool           fun_has_handler; // false if no handlers in fun
  int            label_types[]; // bitvec set if put type on label
  bool           ebx_used;
  bool           esi_used;
  bool           edi_used;
  int            num_callee_used;

  // this stuff changes once per block
  int                 il_block; // implies current handler too
  <instruction>xarray current_instrs; // new one for each block

  // this stuff changes as we walk forward through a block
  int                      il_inst;
  <int, cf_refinement>dict refined_locals;
  <int, *(int[],int)>dict  partial_inits;
  <int>Opt          ebx_local; // unnecessary once we have a supertype
  <int>Opt          esi_local;
  <int>Opt          edi_local;
     // some day put array bounds stuff here
  <*(id,<con>Opt,<instruction>xarray)>xarray other_blocks; // for jump trees
     // may need some web stuff here to help operand2genop!

  // this stuff is what we're building (note that code_blocks are printed
  // as they're created for space considerations)
  <*(id,con)> xarray exports;

   // Cyclone +

  <int,cf_template>dict regions; // Last template dumped in region.
  <int,<int>list>dict   filled; // For each template pointer, 
  // which holes are filled.
  cf_template current_template; // May be null

  Tal::con hole_cons[]; // Pre-condition for each hole.

  // information tracked across functions.
  <string,BITVEC>dict jcc_holes; // function name --> bitvector(1 = jcc,0 = jmp)

  <cf_template,Tal::con>dict template_cons; // TAL con for each template.
  // Cyclone -
} 

extern union tal_place {
  reg Reg;
  int Stackslot;
}

extern bool in_handler     (cg_env env);
extern bool is_handler     (cg_env env, int blockNum);
extern bool handler_changes(cg_env env, int block_num);

extern cg_env new_cg_env (cf_function fn, <*(id,con)>xarray exports);

// environment used for type translation
extern struct tg_env {
  <*(id,kind,int_con_def)>xarray export_cons;
  <*(id,kind,int_con_def)>xarray import_cons;
  <*(id,kind,con)>        xarray file_cons;
  <*(id,con)>             xarray file_abbrevs;
}

// environment used for exception translation
extern struct dg_env {
  <data_block>xarray blocks; 
  <*(id,con)> xarray imports;
  <*(id,con)> xarray exports;
}

#undef list
#undef dict
#undef xarray
#undef id

}}}}}
#endif
