(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Cogen_con:  Constructor transformation and utilities for Code Generation *)

open Utilities
open Tal
open Ycon
open Variable

(* number of bytes in machine word/float *)
val wordsize : int
val floatsize : int

(* Utilities *)
(* normalize ycalculus constructors *)
val ynorm : (variable, ycon) Dict.dict -> ycon -> ycon
(* size of ycalculus constructors *)
val ysizeof : (variable, ykind) Dict.dict -> ycon -> int
(* Get stack information.
   From a normalized ycalculus function type,  Return:
   (tc, (retaddrc, ssz, sigma, rho1, rho2))
     tc: the tal function type
     taldelta: the abstracted type variables
     retaddrc: the return address
*)
val tal_function_type : (variable, ykind) Dict.dict -> ycon -> 
  con * (variable * variable) * con

(* Code Generation *)
(* Generate Tal kinds *)
val cogen_kind : ykind -> kind
(* Generate Tal Exception Stack Type *)
val cogen_exn_stack : variable -> con
(* Generate Tal constructor *)
val cogen_con : ycon -> con
val cogen_cons : ycon list -> con list

(* Module type definitions *)
val toplevel_typedefs : unit -> typedef vector
(* Common Abbreviations *)
val exn_stack_abbrev : variable -> con
val retaddr_abbrev : variable -> variable -> con -> con
val one_arg_abbrev : con -> con -> con
val two_arg_abbrev : con -> con -> con -> con
val multi_arg_abbrev : con list -> con -> con
(*
val one_arg_poly_abbrev : (variable * kind) list -> con -> con -> con
val two_arg_poly_abbrev : (variable * kind) list -> con -> con -> con -> con
val general_fun_abbrev : (variable * kind) list -> con list -> con -> con
*)
