(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Cogen_error:  Error Handling Routines for Code Generation *)

open Format 
open Talpp
open Tal

exception Cogen of string
exception Cogen_unimplemented of string

let die s = (print_string "Cogen error: ";print_string s;print_string"\n";
             raise (Cogen s))

let debug s = (open_vbox 0; print_string "\n"; print_string s; 
               print_string "\n"; 
               print_flush ())

let handle e message =
  try e () with
    x -> debug message; raise x


let dumpye e = (open_vbox 0; print_newline (); 
               Ypp.print_term e; print_newline (); close_box ())

let talppopt = 
  { Talpp.style = Talpp.MASM; 
    Talpp.kinds = true; 
    Talpp.cons = true}

let dumptc c = Talpp.print_con Format.std_formatter talppopt c

let dumptis is = 
  open_hbox ();
  List.map
    (fun i ->
      Talpp.print_instruction Format.std_formatter talppopt i; force_newline())
    is;
  close_box ()

let dumptgen genop = open_hbox ();
  Talpp.print_genop Format.std_formatter talppopt genop;
  print_newline(); close_box()
