#include "unixsupport.h"
#include <sys/mman.h>

/* This version of mmap elides the protection flags portions of mmap.
   We'll just implement read-only, shared pages for now.  Note that
   this will cause a memory protection fault if the returned memory
   is written to.  Really, we should be able to use the TAL type
   system to enable read-only strings and prevent this from
   happening.

   Not bothering with flags other than MAP_SHARED, at least for now.

   We could also think about associating a finalizer with the memory
   allocated here, to do the unmap automatically. */

string unix_mmap(int fd, int offset, int length) {
  void *addr;
  string retval;

  addr = mmap(0, length, PROT_READ, MAP_SHARED, fd, offset);
  if ((int)addr == -1)
    unix_error(__FILE__,__LINE__,"mmap");

  /* create a string with the mapped address as its contents */
  retval = xalloc(sizeof(struct str_internal));
  retval->size = length;
  retval->chars = (char *)addr;

  return retval;
}

void unix_munmap(string s) {
  int retcode;

  if (s == NULL)
    nullpointer_exn(__FILE__,__LINE__);

  retcode = munmap(s->chars,s->size);
  if (retcode == -1)
    unix_error(__FILE__,__LINE__,"munmap");

  /* have to null out the string so we don't segfault in trying
     to read it later */
  s->chars = NULL;
  s->size = 0;
}
