/* Implementation of TAL LOAD macro and popcorn wrapper */

#include <assert.h>
#ifdef TIMEIT
#include <sys/time.h>
#include <unistd.h>
#endif

#include "obj.h"
#include "util.h"

/* to install a .o file in our address space & lookup symbols */
extern struct obj_file *load(char *buf, int buflen);
extern Addr lookup_symbol(struct obj_file *f, const char *symname);

/* for typechecking */
struct string_list {
  char *str;
  struct string_list *next;
};

extern struct string_list *
verify_tal_file (char *typrep, int typereplen,
		 char *tofile, int tofilelen,
		 char *codefile, int codefilelen);
extern int
eqconrep (char *trep1, int rep1len,
	  char *trep2, int rep2len);
      
typedef struct str_internal {int size; char *chars;} *string;
typedef struct rep_internal {int size; char chars[1];} *rep;

/* This is the implementation of the TAL LOAD macro */
Addr *tal_load(string objfilebuf, string tofilebuf, rep typerep) {
  
  char *Cobjfilebuf;
  int Cobjfilelen;
  char *Ctofilebuf; 
  int Ctofilelen;
  char *Ctyperepbuf; 
  int Ctypereplen;
  struct string_list *symbols;
#ifdef TIMEIT
  struct timeval b,a;
  long delta_sec, delta_usec;
#endif

  Cobjfilebuf = objfilebuf->chars;
  Cobjfilelen = objfilebuf->size;
  Ctofilebuf = tofilebuf->chars;
  Ctofilelen = tofilebuf->size;
  Ctyperepbuf = typerep->chars;
  Ctypereplen =  typerep->size;

  /* verify the file first */
  if ((symbols = verify_tal_file (Ctyperepbuf, Ctypereplen,
				  Ctofilebuf, Ctofilelen,
				  Cobjfilebuf, Cobjfilelen))) {
    
  
    int i, num_syms;
    struct string_list *tmp = symbols;
    Addr *symbol_tuple;
    Addr *ret_tuple;
    struct obj_file *f;
    
#ifdef TIMEIT
  gettimeofday(&b,NULL);
#endif
    /* then load it in the address space */
    if ((f = load(Cobjfilebuf,Cobjfilelen))) {
#ifdef TIMEIT
      gettimeofday(&a,NULL);
      delta_sec = a.tv_sec - b.tv_sec;
      delta_usec = a.tv_usec - b.tv_usec;
      if (delta_usec < 0) {
        delta_sec--; delta_usec += 1000000;
      }
      printf ("elapsed time for load was %d.%06d secs\n",
	      delta_sec,delta_usec);
#endif

      /* construct a Popcorn tuple of all of the exported symbols;
         it is important that the order of symbols in this tuple
         matches the order indicated by the type of the caller */
      num_syms = 1;
      while ((tmp = tmp->next)) num_syms++;
      symbol_tuple = (Addr *)xmalloc(sizeof(Addr) * num_syms);

      for (tmp = symbols, i = 0; tmp != NULL; i++, tmp = tmp->next) {
	symbol_tuple[i] = lookup_symbol(f,tmp->str);
	assert(symbol_tuple[i] != (Addr)0);
      }
      ret_tuple = (Addr *)xmalloc(sizeof(Addr));
      *ret_tuple = (Addr)symbol_tuple;
      return ret_tuple;
    }
    else {
      error("Failed to load: malformed object file\n");
      return NULL;
    }
  }
  else {
    error("Failed to load: verification failed\n");
    return NULL;
  }
}

/* This is the implementation of the TAL CAST macro */
void **tal_cast(void* obj, rep trep1, rep trep2) {  
  char *Ctrep1buf;
  int Ctrep1len;
  char *Ctrep2buf; 
  int Ctrep2len;
  
  Ctrep1buf = trep1->chars;
  Ctrep1len = trep1->size;
  Ctrep2buf = trep2->chars;
  Ctrep2len = trep2->size;
  
  /* compare the types */
  if (eqconrep (Ctrep1buf, Ctrep1len, Ctrep2buf, Ctrep2len)) {
    void ** opt = (void **)xmalloc(sizeof(void*));
    opt[0] = obj;
    return opt;
  }
  else {
    return NULL;
  }
}
/*** CODE BELOW GENERATED AUTOMATICALLY BY genCinit.exe ****/



/*** tal_load ***/
typedef struct { int len; char s[147]; } rep147;
static rep147 ctx_tal_load = { 147, { 11,97,0,115,49,0,101,49,0,101,50,0,63,115,122,0,9,52,83,67,73,77,84,86,9,83,3,70,5,82,49,70,9,66,118,3,11,42,5,7,13,33,1,1,15,0,69,9,7,1,17,86,1,82,99,21,42,3,23,33,1,1,25,0,86,3,58,27,29,58,19,31,58,19,33,115,35,70,21,66,42,3,39,33,3,0,1,41,0,86,5,86,7,106,5,45,47,109,1,37,65,43,0,49,67,51,58,53,35,115,55,109,1,57,0,49,67,59,65,7,5,61,65,5,5,63,65,3,3,65,65,1,1,67,70,69,66,42,3,71,33,1,1,73,0,1,75 }};
static struct str_internal str_tal_load = { 8, "tal_load" };

/*** tal_cast ***/
typedef struct { int len; char s[132]; } rep132;
static rep132 ctx_tal_cast = { 132, { 11,97,0,98,0,115,49,0,101,49,0,101,50,0,7,52,83,67,67,86,1,82,99,1,42,3,3,33,1,1,5,0,86,3,82,99,9,42,3,11,33,1,1,13,0,86,5,58,15,17,58,7,19,58,1,21,115,23,70,9,66,42,3,27,33,3,0,1,29,0,86,7,86,9,106,5,33,35,109,1,25,65,31,0,37,67,39,58,41,23,115,43,109,1,45,0,37,67,47,65,9,5,49,65,7,5,51,65,5,3,53,65,3,1,55,65,1,1,57,70,59,66,42,3,61,33,1,1,63,0,1,65 }};
static struct str_internal str_tal_cast = { 8, "tal_cast" };
static void (*add_sym)(string sym,rep r,void *p);
void dyninit_loader(void *unused,void (*f)(string sym,rep r,void *p)) {
  add_sym = f;
  add_sym((string)&str_tal_load,(rep)&ctx_tal_load,(void *) tal_load);
  add_sym((string)&str_tal_cast,(rep)&ctx_tal_cast,(void *) tal_cast);
}
