/* Popcorn Language Stuff
 *
 * Copyright Greg Morrisett, Neal Glew, Dan Grossman
 *   January 1999, all rights reserved
 *
 */

/* Now hacked by Dan and Fred for more informative debugging */

#include <stdlib.h>
#include <stdio.h>

extern void **tal_args;

extern void *GC_malloc(int);
extern void GC_free(void *);

typedef char *Cstring;
typedef struct str_internal {int size; char *chars;} *string;
Cstring string_to_Cstring(string);
extern FILE* tal_stderr;

extern void fprint_string(FILE *,string);
extern void fprint_int(FILE *,int);

/* Call stack related externs */

typedef struct cstk_struct { string hd; struct cstk_struct *tl; } cstk;

/* Type cstk is defined in prelude.tal
   Exn variable __zzzz_active_exn is defined in prelude.tal */
cstk *__zzzz_global_cstk;
cstk *__zzzz_active_cstk;


static void cstk_dump() {
  cstk *x = __zzzz_active_cstk;
  string current;
  int counter;
  fprintf(stderr, "Call stack (only instrumented functions).\n");
  while(x!=NULL) {
    current = x->hd;
    counter=0;
    x=x->tl;
    while(x!=NULL) {
      if(current == x->hd) counter++;
      else break;
      x=x->tl;
    }

    if(counter>0) fprintf(stderr,"(%d)\t",counter+1);
    else fprintf(stderr,"\t");
    fprint_string(tal_stderr,current);
    fprintf(stderr,"\n");
  }

  fprintf(stderr,".\n");
}

void pop_never_null()
{
  fprintf(stderr, "Null pointer exception. \n\n");
  exit(255);
}

void pop_debug_never_null(int i,string s) {
  fprintf(stderr,"Null pointer exception: ");
  fprint_string(tal_stderr,s);
  fprintf(stderr,", char ");
  fprint_int(tal_stderr,i);
  exit(255);
}

typedef struct {
  void *id;
  string info;
  void *value;
} * exn_ptr;

void pop_exn_handler(exn_ptr exn) {
  fprintf(stderr, "Uncaught exception: %p\n",exn);
  fprintf(stderr, "Originating at: ");
  fprint_string(tal_stderr,exn->info);
  fprintf(stderr, "\n");

  if(__zzzz_active_cstk != NULL) cstk_dump();

  exit(255);
}

void pop_exn_handler_verbose(void* exn)
{
  /*  string msg; */

 fprintf(stderr, "Uncaught exception: %p\n",exn);
 // callStackDump();

 fprintf(stderr, "Recovering exception name and location.\n");
 fprintf(stderr, "May cause a crash if code not compiled with verbose exceptions.\n\n");

 fflush(stderr);

 /* Temporarily disabled. -- changing memory format of exceptions.
 msg = (((exn_ptr) exn)->packet->s);

 fprint_string(tal_stderr,msg);
 fprintf(stderr, "\n");
 */

 exit(255);
}

void array_bounds_error(void)
{
  fprintf(stderr, "\narray bounds error!\n");
  exit(255);
}

void division_by_zero_error(void)
{
  fprintf(stderr, "\ndivision by zero error!\n");
  exit(255);
}

/**************************************************************************/

/* Symbol registration for "trusted" symbols needed by dynamic loading */

#ifdef __linux__
#define SYMNAME(s) s
#else
#define SYMNAME(s) "_" ## s
#endif

/* symbols to register from the prelude */
extern long ArrayBounds_pkt;
extern long NullPointer_pkt;
extern long UnionVariant_pkt;
extern long __zzzz_active_exn_pkt;

void 
pop_runtime_init_loader_syms(void (*register_sym)(long,long))
{
  register_sym((long)pop_exn_handler,(long)SYMNAME("pop_exn_handler"));
  register_sym((long)pop_never_null,(long)SYMNAME("pop_never_null"));
  register_sym((long)array_bounds_error,(long)SYMNAME("array_bounds_error"));
  register_sym((long)division_by_zero_error,
	       (long)SYMNAME("division_by_zero_error"));
  register_sym((long)&__zzzz_global_cstk,(long)SYMNAME("__zzzz_global_cstk"));
  register_sym((long)&__zzzz_active_cstk,(long)SYMNAME("__zzzz_active_cstk"));
  register_sym((long)cstk_dump,(long)SYMNAME("cstk_dump"));

  /* stuff from prelude */
  register_sym((long)&ArrayBounds_pkt,(long)SYMNAME("ArrayBounds_pkt"));
  register_sym((long)&NullPointer_pkt,(long)SYMNAME("NullPointer_pkt"));
  register_sym((long)&UnionVariant_pkt,(long)SYMNAME("UnionVariant_pkt"));
  register_sym((long)&__zzzz_active_exn_pkt,
	       (long)SYMNAME("__zzzz_active_exn_pkt"));
}

/*** CODE BELOW GENERATED AUTOMATICALLY BY genCinit.exe ****/

typedef struct rep_internal {int size; char chars[1];} *rep;


/*** __zzzz_global_cstk ***/
typedef struct { int len; char s[42]; } rep42;
static rep42 ctx___zzzz_global_cstk = { 42, { 3,95,63,99,115,116,107,95,116,121,112,101,63,0,1,15,76,1,70,1,66,42,3,3,33,1,1,5,0,70,7,66,42,3,9,33,1,1,11,0,1,13 }};
static struct str_internal str___zzzz_global_cstk = { 18, "__zzzz_global_cstk" };

/*** __zzzz_active_cstk ***/
static rep42 ctx___zzzz_active_cstk = { 42, { 3,95,63,99,115,116,107,95,116,121,112,101,63,0,1,15,76,1,70,1,66,42,3,3,33,1,1,5,0,70,7,66,42,3,9,33,1,1,11,0,1,13 }};
static struct str_internal str___zzzz_active_cstk = { 18, "__zzzz_active_cstk" };

/*** pop_never_null ***/
typedef struct { int len; char s[24]; } rep24;
static rep24 ctx_pop_never_null = { 24, { 1,1,13,99,1,109,3,0,1,67,3,70,5,66,42,3,7,33,1,1,9,0,1,11 }};
static struct str_internal str_pop_never_null = { 14, "pop_never_null" };

/*** pop_debug_never_null ***/
typedef struct { int len; char s[91]; } rep91;
static rep91 ctx_pop_debug_never_null = { 91, { 5,115,0,63,115,122,0,5,83,73,53,52,84,86,3,83,5,70,7,82,49,70,11,66,118,5,13,42,3,15,33,1,1,17,0,70,19,82,42,5,9,21,33,1,1,23,0,69,3,3,3,25,86,1,58,27,29,58,1,31,58,1,33,115,35,99,1,109,1,37,0,39,67,41,65,1,1,43,70,45,66,42,3,47,33,1,1,49,0,1,51 }};
static struct str_internal str_pop_debug_never_null = { 20, "pop_debug_never_null" };

/*** pop_exn_handler ***/
typedef struct { int len; char s[128]; } rep128;
static rep128 ctx_pop_exn_handler = { 128, { 5,99,36,48,0,63,115,122,0,5,77,73,71,84,74,4,42,3,3,86,1,33,1,1,5,1,7,66,70,9,82,86,3,83,13,70,15,82,49,70,19,66,118,13,21,42,3,23,33,1,1,25,0,70,27,82,42,5,17,29,33,1,1,31,0,69,3,3,1,33,70,35,82,42,7,11,37,7,33,1,1,39,0,69,1,1,1,41,101,58,43,45,115,47,99,1,109,1,49,0,51,67,53,58,55,47,115,57,109,1,59,0,51,67,61,70,63,66,42,3,65,33,1,1,67,0,1,69 }};
static struct str_internal str_pop_exn_handler = { 15, "pop_exn_handler" };

/*** pop_exn_handler_verbose ***/
static rep128 ctx_pop_exn_handler_verbose = { 128, { 5,99,36,48,0,63,115,122,0,5,77,73,71,84,74,4,42,3,3,86,1,33,1,1,5,1,7,66,70,9,82,86,3,83,13,70,15,82,49,70,19,66,118,13,21,42,3,23,33,1,1,25,0,70,27,82,42,5,17,29,33,1,1,31,0,69,3,3,1,33,70,35,82,42,7,11,37,7,33,1,1,39,0,69,1,1,1,41,101,58,43,45,115,47,99,1,109,1,49,0,51,67,53,58,55,47,115,57,109,1,59,0,51,67,61,70,63,66,42,3,65,33,1,1,67,0,1,69 }};
static struct str_internal str_pop_exn_handler_verbose = { 23, "pop_exn_handler_verbose" };

/*** array_bounds_error ***/
static rep24 ctx_array_bounds_error = { 24, { 1,1,13,99,1,109,3,0,1,67,3,70,5,66,42,3,7,33,1,1,9,0,1,11 }};
static struct str_internal str_array_bounds_error = { 18, "array_bounds_error" };

/*** division_by_zero_error ***/
static rep24 ctx_division_by_zero_error = { 24, { 1,1,13,99,1,109,3,0,1,67,3,70,5,66,42,3,7,33,1,1,9,0,1,11 }};
static struct str_internal str_division_by_zero_error = { 22, "division_by_zero_error" };

/*** GC_free ***/
typedef struct { int len; char s[152]; } rep152;
static rep152 ctx_GC_free = { 152, { 17,97,0,110,0,101,0,97,101,98,120,0,97,101,115,105,0,97,101,100,105,0,97,101,98,112,0,115,49,0,11,77,78,67,52,83,63,86,3,78,1,86,15,58,3,5,115,7,86,13,86,7,86,9,86,11,86,5,109,0,9,11,66,13,115,15,100,17,0,19,67,21,58,23,7,115,25,86,1,33,1,1,29,0,99,3,3,85,31,106,5,33,19,109,0,27,11,66,13,115,15,100,17,0,35,67,37,65,15,9,39,65,13,7,41,65,11,7,43,65,9,7,45,65,7,7,47,65,5,5,49,65,3,3,51,65,1,1,53,70,55,66,42,3,57,33,1,1,59,0,1,61 }};
static struct str_internal str_GC_free = { 7, "GC_free" };
static void (*add_sym)(string sym,rep r,void *p);
void dyninit_pop_runtime(void *unused,void (*f)(string sym,rep r,void *p),
			 void *unused2) {
  add_sym = f;
  add_sym((string)&str___zzzz_global_cstk,(rep)&ctx___zzzz_global_cstk,(void *)&__zzzz_global_cstk);
  add_sym((string)&str___zzzz_active_cstk,(rep)&ctx___zzzz_active_cstk,(void *)&__zzzz_active_cstk);
  add_sym((string)&str_pop_never_null,(rep)&ctx_pop_never_null,(void *)&pop_never_null);
  add_sym((string)&str_pop_debug_never_null,(rep)&ctx_pop_debug_never_null,(void *) pop_debug_never_null);
  add_sym((string)&str_pop_exn_handler,(rep)&ctx_pop_exn_handler,(void *)&pop_exn_handler);
  add_sym((string)&str_pop_exn_handler_verbose,(rep)&ctx_pop_exn_handler_verbose,(void *)&pop_exn_handler_verbose);
  add_sym((string)&str_array_bounds_error,(rep)&ctx_array_bounds_error,(void *)&array_bounds_error);
  add_sym((string)&str_division_by_zero_error,(rep)&ctx_division_by_zero_error,(void *)&division_by_zero_error);
  add_sym((string)&str_GC_free,(rep)&ctx_GC_free,(void *) GC_free);
}
